unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Masks, ComCtrls, ExtCtrls;

type
  TForm1 = class(TForm)
    Button1: TButton;
    Button2: TButton;
    Memo1: TMemo;
    Memo2: TMemo;
    Label1: TLabel;
    StatusBar1: TStatusBar;
    Timer1: TTimer;
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
  private
    cancelled:boolean;
    masks:array of TMask;
    total, scanned:integer;
    procedure startscan(sender:tobject; filename:string);
    procedure finishscan(sender:tobject; filename:string);
    procedure abortscan(sender:tobject; filename:string);
    procedure virusfound(sender:tobject; filename:string);
    procedure error(sender:tobject; Filename:string; error:exception);
    function shouldScan(s:string):boolean;
    procedure scanMasks(s:string);
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

uses SyncObjs;

type
  TScanFileEvent=procedure(Sender:TObject; Filename:string) of object;
  TScanErrorEvent=procedure(Sender:TObject; Filename:string; Error:Exception) of object;

  TScanningThread=class(TThread)
  private
    FCS:TCriticalSection;
    FIndex:integer;
    fn:string;
    err:Exception;
    FList:TStringList;
    FOnFinishedScanFile: TScanFileEvent;
    FOnAbortedScanFile: TScanFileEvent;
    FOnStartScanFile: TScanFileEvent;
    FOnVirusFound: TScanFileEvent;
    FOnError: TScanErrorEvent;
    procedure DoStartScan;
    procedure DoFinishScan;
    procedure DoAbortScan;
    procedure DoVirusFound;
    procedure DoError;
    function GetThreadName: string;
  protected
    procedure execute; override;
  public
    constructor Create(AIndex:integer);
    destructor Destroy; override;
    procedure scan(AFilename:string);
    function IsScannig:boolean;
    property OnStartScanFile:TScanFileEvent read FOnStartScanFile write FOnStartScanFile;
    property OnFinishedScanFile:TScanFileEvent read FOnFinishedScanFile write FOnFinishedScanFile;
    property OnAbortedScanFile:TScanFileEvent read FOnAbortedScanFile write FOnAbortedScanFile;
    property OnVirusFound:TScanFileEvent read FOnVirusFound write FOnVirusFound;
    property OnError:TScanErrorEvent read FOnError write FOnError;
    property ThreadName:string read GetThreadName;
    property FileName:string read fn;
  end;

var cs:TCriticalSection=nil;

procedure logtofile(s:string);
var f:textfile;
begin
  try
    cs.Acquire;
    try
      assignfile(f, 'error.log');
      try
        append(f);
      except
        rewrite(f);
      end;
      try
        writeln(s);
      finally
        closefile(f);
      end;
    finally
      cs.Release;
    end;
  except
    // this procedure must not error
  end;
end;

{ TScanningThread }

constructor TScanningThread.Create(AIndex:integer);
begin
  FCS:=TCriticalSection.Create;
  FList:=TStringList.Create;
  FIndex:=AIndex;
  inherited Create(true);
end;

destructor TScanningThread.Destroy;
begin
  freeandnil(FList);
  freeandnil(FCS);
  inherited;
end;

procedure TScanningThread.DoAbortScan;
begin
  if assigned(FOnAbortedScanFile) then
    FOnAbortedScanFile(self, fn);
end;

procedure TScanningThread.DoError;
begin
  if assigned(FOnError) then
    FOnError(self, fn, err);
end;

procedure TScanningThread.DoFinishScan;
begin
  if assigned(OnFinishedScanFile) then
    OnFinishedScanFile(self, fn);
end;

procedure TScanningThread.DoStartScan;
begin
  if assigned(OnStartScanFile) then
    OnStartScanFile(self, fn);
end;

procedure TScanningThread.DoVirusFound;
begin
  if assigned(OnVirusFound) then
    OnVirusFound(self, fn);
end;

procedure TScanningThread.execute;
var f:TFileStream;
    s,t:string;
    count:integer;
    i,j,l:integer;
begin
  setlength(s, 64000);
  t:='virus';// what to search for
  l:=length(t);
  while not terminated do
  try
    fn:='';
    FCS.Acquire;// extract first file if any (FIFO)
    try
      if FList.Count>0 then
      begin
        fn:=FList[0];
        FList.Delete(0);
      end;
    finally
      FCS.Release;
    end;
    if fn='' then// no file? go to sleep and get next file when woken up
    begin
      suspend;
      continue;
    end;
    f:=TFileStream.Create(fn, fmopenread);
    try
      Synchronize(DoStartScan);
      while (not terminated) and (f.Position<f.Size) do
      begin
        count:=f.Read(pchar(s)^, length(s)-1);
        s[count+1]:=#0;
        i:=1;
        j:=1;
        while (i<=count-l+1) and (j<=l) do
        begin
          j:=1;
          while (j<=l) and (s[i]=t[j]) do
          begin
            inc(j);
            inc(i);
          end;
          i:=i-j+2;
        end;
        if j>l then
          Synchronize(DoVirusFound);
      end;
      if Terminated then
        Synchronize(DoAbortScan)
      else
        Synchronize(DoFinishScan);
    finally
      freeandnil(f);
    end;
  except
    on e:exception do
    begin
      err:=e;
      Synchronize(DoError);
    end;
  end;
end;

function TScanningThread.GetThreadName: string;
begin
  result:='Thread '+inttostr(FIndex);
end;

function TScanningThread.IsScannig: boolean;
begin
  FCS.Acquire;
  try
    result:=(not Suspended) or (FList.Count>0);
  finally
    FCS.Release;
  end;
end;

procedure TScanningThread.scan(AFilename: string);
begin
  FCS.Acquire;
  try
    FList.Add(AFilename);
  finally
    FCS.Release;
  end;
  if Suspended then
    Resume;// wake up the thread if it's sleeping
end;

procedure TForm1.abortscan(sender: tobject; filename: string);
begin
  memo1.lines.add((sender as TScanningThread).ThreadName+' aborted while scanning file '+(sender as TScanningThread).Filename);
  inc(scanned);
end;

procedure TForm1.Button1Click(Sender: TObject);
const max=10;
var threads:array[0..max-1] of TScanningThread;
    i:integer;
  procedure findfiles(path:string);
  var r:tsearchrec;
  begin
    path:=IncludeTrailingPathDelimiter(path);
    if findfirst(path+'*.*', faanyfile-favolumeid, r)=0 then
    begin
      repeat
        application.processmessages;// make sure GUI doesn't freeze;
        if r.Attr and fadirectory=fadirectory then
        begin
          if (r.Name<>'.') and (r.Name<>'..') then
            findfiles(path+r.name);
        end                                   else
        begin
          if shouldScan(path+r.name) then
          begin
            inc(total);
            threads[i].scan(path+r.name);
            i:=(i+1) mod max;// add each file to the next thread. when end of list, go the the first one
          end;
        end;
      until cancelled or (findnext(r)<>0);
      findclose(r);
    end;
  end;
begin
  timer1.enabled:=true;
  statusbar1.simpletext:='searching for files';
  total:=0;
  scanned:=0;
  cancelled:=false;
  memo1.Clear;
  memo2.Clear;
// user clicked scan
  for i:=1 to max do
  begin
    threads[i-1]:=TScanningThread.create(i);
    threads[i-1].OnStartScanFile:=startscan;
    threads[i-1].OnFinishedScanFile:=finishscan;
    threads[i-1].OnAbortedScanFile:=abortscan;
    threads[i-1].OnVirusFound:=virusFound;
    threads[i-1].OnError:=error;
  end;
  try
    i:=0;
    findfiles('c:\1');
    statusbar1.simpletext:='waiting for threads';
    // after finished searching for all files, wait for threads to finish
    repeat
      i:=0;
      while (i<max) and (not threads[i].IsScannig) do
        inc(i);
      sleep(10);// do not kill the cpu
      application.processmessages;// don't freeze the GUI
    until (i=max) or cancelled;// cancelled is a boolean value which is set to true when you click a cancel button
    if cancelled then
      memo1.lines.add('Scan aborted')
    else
      memo1.lines.add('Scan finished');
  finally
    statusbar1.simpletext:='finished';
    timer1.enabled:=false;
    for i:=1 to max do// terminate and destroy all threads
    begin
      threads[i-1].Terminate;
      threads[i-1].Resume;// just in case it's sleeping
      threads[i-1].WaitFor;
      application.processmessages;// waitfor might take a little time so do not freeze the GUI
    end;
  end;
end;

procedure TForm1.Button2Click(Sender: TObject);
begin
  cancelled:=true;
end;

procedure TForm1.finishscan(sender: tobject; filename: string);
begin
  memo1.lines.add((sender as TScanningThread).ThreadName+' finished scanning file '+(sender as TScanningThread).Filename);
  inc(scanned);
end;

function TForm1.shouldScan(s: string): boolean;
var i:integer;
begin
  i:=0;
  while (i<length(masks)) and (not masks[i].matches(s)) do
    inc(i);
  result:=i<length(masks);
end;

procedure TForm1.startscan(sender: tobject; filename: string);
begin
  memo1.lines.add((sender as TScanningThread).ThreadName+' started scannig file '+(sender as TScanningThread).Filename);
end;

procedure TForm1.virusfound(sender: tobject; filename: string);
begin
  memo2.lines.add('found virus in file '+filename);
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  scanMasks('*.exe,*.txt, *.doc, *.pdf');
end;

procedure TForm1.FormDestroy(Sender: TObject);
var i:integer;
begin
  for i:=1 to length(masks) do
    freeandnil(masks[i-1]);
  setlength(masks, 0);
end;

procedure TForm1.error(sender: tobject; filename:string; error: exception);
var s:string;
begin
  s:='** Error while scannig file "'+filename+'": '+error.message;
  logtofile(s);
  memo1.lines.add(s);
  inc(scanned);
end;

procedure TForm1.Timer1Timer(Sender: TObject);
begin
  label1.caption:=format('Total: %d | Scanned: %d', [total, scanned]);
end;

procedure TForm1.scanMasks(s: string);
var l:TStringList;
    i:integer;
begin
  l:=TStringList.Create;
  try
    l.CommaText:=s;
    setlength(masks, l.count);
    for i:=1 to l.count do
      masks[i-1]:=TMask.Create(l[i-1]);
  finally
    freeandnil(l);
  end;
end;

initialization
  cs:=TCriticalSection.create;

finalization
  freeandnil(cs);

end.
